from __future__ import absolute_import

from badook_tests import *
from badook_tests.data_test_case import DataTestCase
from badook_tests.runner import BadookRuntimeTestResult, BadookBaseTestResult, BadookTestRunnerError

__version__ = '0.2.0-pre.0'



import click
import os
import shutil
import inspect


# validation on project name - return bad_parameter_error if project_name contains space
def project_name_validation(ctx, param, value):
    try:
        project_name = value
        if ' ' not in project_name:
            return project_name
        else:
            raise click.BadParameter(
                'The project_name need to be without spaces')
    except ValueError:
        click.echo('Incorrect project_name: {}'.format(project_name))
        value = click.prompt(param.prompt)
        return project_name_validation(ctx, param, value)


@click.command()
# ask for root to folder
@click.option('--root', '-r',  help='Specify root of project''s folder', default='.', prompt="folder root")
# ask for Project name
@click.option('--project_name', '-n', prompt="Project_name",
              help='Specify a project name without spaces', callback=project_name_validation)
# ask for root to api_id
@click.option('--api_id', '-id',  prompt=True, help='The user API_ID. '
                                                    'See badook''s web application-->Settings for further info')
# ask for root to api_key
@click.option('--api_key', prompt=True, hide_input=True, help='The user API_Key for authentication'
                                                              'See badook''s web application-->Settings for further info')
@click.option('--base_uri', '-u',
              help='Specify uri for runtime badook. defult set to https://runtime.badook.ai/services/',
              default='https://runtime.badook.ai/services/')
def Process(project_name, root, api_id, api_key, base_uri):
    click.echo('Project set to {}'.format(project_name))
    # Define folders path by root + project name
    path = f"{root}/{project_name}"
    if not os.path.exists(path):
        """Create folders according to root tree:
        root
        | -> tests /
        | | -> test_example.py
        | ->.badook /
        | | -> config / default.yaml
        | ->.gitignore
        """
        folders = {"/tests", "/.badook", "/.gitignore"}
        for folder in folders:
            os.makedirs(path+folder)
        # create test_example.py with
        with open(path+'/tests/test_example.py', 'w') as f:
            # Summary and tests code exmaples
            text = """
            #This file containes stractures of summaries, tests and assertions built with badook-ai

            from badook_tests.test import TestCase
            class RealEstateTestCase(TestCase):
            def set_up(self):
            # setup dataset and feature level summaries
            def test_feature_distribution(self):

            # Summaries
            def set_up(self):
            context = BadookContext(client_id="", client_secret="")

            ## Defining a dataset level summary
            self.real_estate_summary = context \
            .from_dataset('realestateData') \
            .set_name('real_estate_summary')

            ## Defining a feature level numeric summary
            type_sub = NumericSummary(feature='price', name='sub_price') \
            .group_by('type', 'suburb')
            .on(Self.real_estate_summary)

            self.summary_builder.add_summary(type_sub)


            #Test containing dataset level checks:
            def test_real_estate_summary_dataset(self):
            self.real_estate_summary.features.check(f => 'price'))
            self.real_estate_summary.records.compare_to(ComparisonPoint.LAST)
            .check(current, past => current.count >= past.count)
            """
            f.write(text)

        # Ask for the user an api_key and user_id and create .yaml file
        os.mkdir(path+'/.badook/config')
        text = inspect.cleandoc("""base_uri: '"""+base_uri+"""'
            api_key:'""" + api_id + """'
            secret_key:'""" + api_key + """'
            project_name:'"""+project_name+"""'
            """)
        # save configuration to .yaml file. 3 last row are indent
        with open(path+'/.badook/config/default.yaml', 'w') as f:
            f.write(text)

        # Create .goignore file in path/gitignore with instruction to ignore the .badook folder
        text_ignore = inspect.cleandoc("""#ignore all files in path/.badook/ folder
                        .badook""")

        with open(path+'/.gitignore/.gitignore', 'w') as f:
            f.write(text_ignore)

        click.echo(f"\nDirectory was created at {path}")

    else:
        # delete the folder structure everytime running this command - for tests only
        # shutil.rmtree(path,ignore_errors=True)
        click.echo(f"\n Directory {project_name} is already exists")


if __name__ == '__main__':
    Process(None, None, None, None, None)



from __future__ import annotations
from badook_tests.dsl.summary import Summary
from badook_tests.dsl.checks import Calculation


class DBSCAN(Summary):
    """
    DBSCAN Clustering: https://en.wikipedia.org/wiki/DBSCAN
    """

    def __init__(self, features: list, name: str, eps=0.5, min_samples=5, metric='euclidean',
                 metric_params=None, algorithm='auto', leaf_size=30, p=None, n_jobs=None):
        super().__init__(features, name)
        self.method = 'dbscan'
        self.type = 'Cluster'
        self.params = {'eps': eps, 'min_samples': min_samples, 'metric': metric,
                       'metric_params': metric_params, 'algorithm': algorithm,
                       'leaf_size': leaf_size, 'p': p, 'n_jobs': n_jobs}

    @property
    def clusters_count(self):
        return Calculation(self.data, 'clusters_count', self._ctx)

    @property
    def noise_count(self):
        return Calculation(self.data, 'noise_count', self._ctx)

    @property
    def silhouette_score(self):
        return Calculation(self.data, 'silhouette_score', self._ctx)

    @property
    def clusters(self):
        return Clusters(self.data, 'cluster', self._ctx)


class Clusters(Summary):
    """
    A subclass to hold the output of the clusters
    """

    def __init__(self, data, name, _ctx):
        self.data = data
        self.name = name
        self._ctx = _ctx

    @property
    def member_count(self):
        return Calculation(self.data, self.name + '_member_count', self._ctx)

    @property
    def core_count(self):
        return Calculation(self.data, self.name + '_membcore_counter_count', self._ctx)

    @property
    def border_count(self):
        return Calculation(self.data, self.name + '_border_count', self._ctx)



import unittest
import click
import sys
from collections import namedtuple


class BadookTestRunnerError(Exception):
    """Raised when there's a problem running a badook test"""

    def __init__(self, message):
        self.msg = message
        super().__init__(self.msg)


class BadookBaseTestResult(unittest.TextTestResult):
    """A base class to communicate the result of a test (a single test case) with the badook lab runtime"""

    result = namedtuple('results', ['passed', 'failed'])
    result.passed = []
    result.failed = []

    def startTest(self, test):

        # if there are no issues with the user's script (e.g. syntax error etc)
        if not isinstance(test, unittest.loader._FailedTest):
            # TODO: move this action to pyspark-runner when the test is finished. And send here just status success or not
            data = {'orgId': test.channel_id, 'projectId': test.context.project_name,
                    'runId': test.run_id, 'functionName': str(test._testMethodName), 'status': 'Running'}

            response = ManagementClient.register_start_test(test, data)

            # if error when requesting to start a test in service manager, abort the test
            if response['error']:
                msg = response['error']['message']
                raise BadookTestRunnerError(msg)

            super().startTest(test)
        else:
            test.__getattr__(test._testMethodName)()

    def addError(self, test, err):
        name = str(test._testMethodName) if hasattr(
            test, '_testMethodName') else ""

        data = {'orgId': test.channel_id if hasattr(test, 'channel_id') else "",
                'projectId': test.context.project_name if hasattr(test, 'context') else "",
                'runId': test.run_id if hasattr(test, 'run_id') else "",
                'functionName': str(test._testMethodName) if hasattr(test, '_testMethodName') else "",
                'status': 'Failure', 'failureMessage': str(err)}

        self.result.failed.append(name)

        # in case the run was aborted before a test was created
        response = {'error': True}
        try:
            # if test exists, update it's status in service management
            if test.context:
                response = test.context.api_client.send_json_post_request(
                    data, "runhistory/test")

                # if error when requesting to update a test in service manager - abort
                if response['error']:
                    msg = response['error']['message']
                    raise BadookTestRunnerError(msg)

        except Exception as e:
            BadookTestRunnerError(e)

        # If test exists:
        # this will display a failed test error message for the user.
        # passing err elements (list) except the traceback: err = (type, value, traceback)
        errNoTraceback = (err[0], err[1], None)
        super().addError(test, errNoTraceback)

    def addSuccess(self, test):
        name = str(test._testMethodName)
        response = ManagementClient.update_success(test, name)

        # if error when requesting to update a test in service manager - abort
        if response['error']:
            msg = response['error']['message']
            raise BadookTestRunnerError(msg)
        self.result.passed.append(name)
        super().addSuccess(test)


class ManagementClient(object):
    '''
    This is a utility class for accessing the management API.
    Its main use is to be mocked in tests. You can patch the ManagementClient methods as follows:

    @patch('badook_tests.runner.ManagementClient.[function name]', return_value=[returned value])
    '''

    @classmethod
    def register_start_test(cls, test, data):
        return test.context.api_client.send_json_post_request(data, "runhistory/test")

    @classmethod
    def update_success(self, test, name):
        data = {'orgId': test.channel_id, 'projectId': test.context.project_name,
                'runId': test.run_id, 'functionName': name, 'status': 'Success'}

        response = test.context.api_client.send_json_post_request(
            data, "runhistory/test")

        return response


class BadookRuntimeTestResult(BadookBaseTestResult):
    """A result class to communicate the result of a test (a single test case) with the badook lab runtime"""

    def startTest(self, test):

        try:
            super().startTest(test)
        except BadookTestRunnerError as berr:
            click.secho(
                f'\nUh Oh! Something bad happened while trying to run the test. Please contact Badook support.\n(Management Services - {berr.msg})', fg='red', bold=True)
            sys.exit(0)
        except Exception as err:
            click.secho(f'\n{err}', fg='red', bold=True)
            sys.exit(0)



from typing import Optional
from .settings import Settings


_config: Settings = None


def set_config(config_file_path: Optional[str]):
    global _config
    _config = Settings(config_file_path=config_file_path)


def set_config_from_settings(settings: Settings):
    global _config
    _config = settings


def get_config():
    return _config



from __future__ import annotations
from .checks import Calculation
from badook_tests.dsl.enums import ComparisonPoint
from typing import Optional
import unittest
import logging
import sys
import json
import functools
import shortuuid


class DatasetSummary(object):

    name = ""
    display_name = ""

    def __init__(self, ctx):
        self._ctx = ctx
        self.summaries = {}
        self.joined_datasets = {}
        self.user_features = {}

    def set_dataset(self, dataset):
        self.dataset = dataset

    def add_summary(self, summary: Summary) -> DatasetSummary:
        self.summaries[summary.name] = summary
        return self

    def set_name(self, name: str) -> DatasetSummary:
        # this is the summary folder's name in storage which needs to be unique
        self.name = f'{name[0:24]}-{shortuuid.uuid()}'

        # this is the summary name itself for display
        self.display_name = name
        return self

    def get_summary(self, name: str) -> Summary:
        return self.summaries[name]

    def collect(self):
        self._ctx.collect(self, 'summaries')

    def set_project_id(self, id):
        self.project_id = id

    def set_run_id(self, id):
        self.run_id = id

    def set_batch(self, batch_value):
        self.batch_value = batch_value

    def join_on(self, name, key: list = None, key_l: list = None, key_r: list = None) -> DatasetSummary:
        keys = []
        # if both datasets share the same key name
        if key:
            for key_item in key:
                keys.append([key_item, key_item])

        # if each dataset has different name for the key
        elif key_l and key_r and len(key_l) == len(key_r):
            for index, ket_l_item in enumerate(key_l, start=0):
                keys.append([ket_l_item, key_r[index]])

        else:
            raise ValueError("Invalid keys")

        self.joined_datasets[name] = {'keys': keys}

        """
        Dict structure:
        for joinOn 1:
            key_temp =  [[<key1A>, <key1B>],[<key2A>, <key2B>]]
            joined_datasets['name1'] = {'keys': key_temp }

        for joinOn 2:
            key_temp =  [[<keyCA>, <keyCA>]]
            joined_datasets['name2'] = {'keys': key_temp }

        The above process creates the following dict:
        joined_dataset = {
            name1:  {
                'keys':[[<key1A>, <key1B>],[<key2A>, <key2B>]],
                'batch_value: <name> -> optional (will be added in test_cases.py)
            },
            name2:  {
                'keys':[[<keyCA>, <keyCA>]],
                'batch_value: <name> -> optional (will be added in test_cases.py)
            }
        }
        """
        return self

    def with_user_feature(self, name, features, expression, feature_type) -> DatasetSummary:
        dict_temp = {'expression': expression,
                     'feature': features, 'type':  feature_type}
        self.user_features[name] = dict_temp
        return self

    def get_user_feature(self, features, name: str) -> DatasetSummary:
        if name in self.user_features:
            return self.user_features[name]
        else:
            raise ValueError(f"user_feature {name} was not set")


class SummaryData(object):

    def __init__(self, dataset, summary_name, name, run_id, join_on, time_window, summary_type, model_params):
        self.name = name
        self.summary_name = summary_name
        self.run_id = run_id
        self.join_on = join_on
        self.time_window = time_window
        self.dataset = dataset
        self.type = summary_type
        self.model_params = model_params


class Summary(object):

    _ctx = None

    def __init__(self, feature: str, name: str):
        self.name = name
        self.feature = feature
        self.user_calculations = {}
        self.feature_set = {}
        self.groups = None
        self.time_window = None
        self.type = ''
        self.model_params = {
            'isFit': False,
            'isPredict': False
        }

    def group_by(self, *args) -> Summary:
        self.groups = list(args)
        return self

    def set_time_window(self, time_key: str, time_format: str,  units: str, number_of_units):
        # ToDo: Change the format option to spark options in the docstring. Change the units options to the runner
        # options in the docstring.
        """
        Time format option: {'date', timestamp'}
        Units options: {years, months, weeks, days, hours, minutes, seconds}
        """
        self.time_window = {
            'time_key': time_key,
            'number_of_units': number_of_units,
            'units': units,
            'time_format': time_format
        }
        return self

    def add_udc(self, name: str, calculation, input_columns=None) -> Summary:
        self.user_calculations[name] = {
            'func': calculation,
            'input_columns': input_columns
        }

    def get_udc(self, name: str) -> Calculation:
        if name in self.user_calculations:
            return self.__dict__[name]
        else:
            raise ValueError(f"udc {name} was not set")

    def on(self, dataset_summary: DatasetSummary) -> Summary:
        dataset_summary.add_summary(self)
        self._ctx = dataset_summary._ctx
        self.dataset = dataset_summary.dataset
        self.parent = dataset_summary.name

        return self

    def get_historical_run(self, offset=None, run_id=None):

        if offset:
            run_id = int(self.run_id) - offset

        summary_folder_name = self.parent
        display_name = self._ctx.dataset_summaries[summary_folder_name]['display_name']

        data = {
            'projectId': self._ctx.project_name,
            'runId': run_id,
            'datasetDisplayName': display_name
        }

        # get offset run's summary data from service managment
        response = self._ctx.api_client.send_json_post_request(
            data, "runhistory/run/summary")
        if response['error']:
            message = f'Failed getting summary from run {run_id}: ' + \
                response['error']['message']
            raise Exception(message)

        return SummaryData(self.dataset, self.name, response['data']['name'], int(run_id), self.groups,
                           self.time_window, self.type, self.model_params)

    def set_run_id(self, id):
        self.run_id = id
        self.data = SummaryData(self.dataset, self.name, self.parent, self.run_id, self.groups,
                                self.time_window, self.type, self.model_params)

        if len(self.user_calculations) > 0:
            # add user calculation as a methods
            for udc in self.user_calculations:
                self.__dict__[udc] = Calculation(self.data, udc, self._ctx)

    def fit(self):
        self.model_params['isFit'] = True
        return self

    def predict(self):
        self.model_params['isPredict'] = True
        return self

    def fit_predict(self):
        self.model_params['isFit'] = True
        self.model_params['isPredict'] = True
        return self


class DurationSummary(Summary):

    def __init__(self, feature: str, name: str, time_format=None):
        super().__init__(feature, name)
        self.type = 'DurationSummary'
        self.feature_format = time_format

    @property
    def min(self):
        return Calculation(self.data, 'min', self._ctx)

    @property
    def max(self):
        return Calculation(self.data, 'max', self._ctx)

    @property
    def duration(self):
        return Calculation(self.data, 'duration', self._ctx)


class CounterSummary(Summary):

    def __init__(self, feature: str, name: str):
        super().__init__(feature, name)
        self.type = 'CounterSummary'

    @property
    def count(self):
        return Calculation(self.data, 'count', self._ctx)

    @property
    def count_distinct(self):
        return Calculation(self.data, 'count_distinct', self._ctx)


class NumericSummary(CounterSummary):

    def __init__(self, feature: str, name: str):
        super().__init__(feature, name)
        self.type = 'NumericSummary'

    @property
    def min(self):
        return Calculation(self.data, 'min', self._ctx)

    @property
    def max(self):
        return Calculation(self.data, 'max', self._ctx)

    @property
    def std_dev(self):
        return Calculation(self.data, 'std_dev', self._ctx)

    @property
    def mean(self):
        return Calculation(self.data, 'mean', self._ctx)

    @property
    def median(self):
        return Calculation(self.data, 'median', self._ctx)

    @property
    def sum(self):
        return Calculation(self.data, 'sum', self._ctx)


class PointSummary(NumericSummary):

    def __init__(self, feature: str, name: str):
        super().__init__(feature, name)
        self.type = 'PointSummary'

    @property
    def count(self):
        return Calculation(self.data, 'count', self._ctx)

    @property
    def count_distinct(self):
        return Calculation(self.data, 'count_destinct', self._ctx)

    @property
    def group(self):
        return Calculation(self.data, 'group', self._ctx)

    @property
    def max(self):
        return Calculation(self.data, 'max', self._ctx)

    @property
    def poly_fit_params(self):
        return Calculation(self.data, 'min', self._ctx)

    @property
    def dimension(self):
        return Calculation(self.data, 'dimension', self._ctx)


class PolygonSummary(NumericSummary):

    def __init__(self, feature: str, name: str):
        super().__init__(feature, name)
        self.type = 'PolygonSummary'

    @property
    def count(self):
        return Calculation(self.data, 'count', self._ctx)

    @property
    def count_distinct(self):
        return Calculation(self.data, 'count_distinct', self._ctx)

    @property
    def mean_area(self):
        return Calculation(self.data, 'mean_area', self._ctx)

    @property
    def max_area(self):
        return Calculation(self.data, 'max_area', self._ctx)

    @property
    def min_area(self):
        return Calculation(self.data, 'min_area', self._ctx)

    @property
    def std_area(self):
        return Calculation(self.data, 'std_area', self._ctx)

    @property
    def union(self):
        return Calculation(self.data, 'union', self._ctx)


class BooleanSummary(CounterSummary):

    def __init__(self, feature: str, name: str):
        super().__init__(feature, name)
        self.type = 'BooleanSummary'

    @property
    def countTrue(self):
        return Calculation(self.data, 'countTrue', self._ctx)

    @property
    def countFalse(self):
        return Calculation(self.data, 'countFalse', self._ctx)


class CategoricalSummary(Summary):

    def __init__(self, feature: str, name: str):
        super().__init__(feature, name)
        self.type = 'CategoricalSummary'

    @property
    def count(self):
        return Calculation(self.data, 'count', self._ctx)

    @property
    def percentage(self):
        return Calculation(self.data, 'percentage', self._ctx)



from typing import Dict, Tuple, Any, Optional
from yaml import safe_load
from pydantic import BaseSettings, HttpUrl, FilePath
from pydantic.env_settings import SettingsSourceCallable


class Settings(BaseSettings):
    '''
    This class defines the configuration structure for the entire application.
    '''
    config_file_path: Optional[FilePath]
    data_cluster_url: HttpUrl
    management_cluster_url: HttpUrl
    client_id: str
    client_secret: str

    class Config:
        '''
        Environment configuration for loading the Settings by pydantic.
        '''
        env_prefix = 'BDK_'

        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            '''
            Add the YAML loading method to the configuration sources
            '''
            config_file_path = init_settings(None)['config_file_path']
            return (
                init_settings,
                env_settings,
                file_secret_settings,
                _yaml_config_setting(config_file_path, True),
                _yaml_config_setting('config/badook.yaml', False),
            )


class _yaml_config_setting:
    def __init__(self, config_file_path: str, throw_if_not_exists: bool) -> None:
        self.config_file_path = config_file_path
        self.throw_if_not_exists = throw_if_not_exists

    def __call__(self, _: BaseSettings) -> Dict[str, Any]:
        '''
        Read settings from a yaml source. Optionally throw an exception if the file does not exist.
        '''
        if self.config_file_path is None:
            return {}

        try:
            with open(self.config_file_path) as config_file:
                return safe_load(config_file)
        except FileNotFoundError:
            if self.throw_if_not_exists:
                raise

            return {}



from .run import run
from .project import project
import emoji
import click
import logging
import sys
from ..config import set_config

logger = logging.getLogger(__name__)
logger.disabled = True
logger.setLevel(logging.ERROR)
ch = logging.StreamHandler(sys.stdout)
logger.addHandler(ch)


@click.group()
@click.option('--config_file_path', '-c', default=None)
def bdk(config_file_path):
    set_config(config_file_path=config_file_path)
    click.secho(emoji.emojize('Welcome to badook cli :falafel:',
                use_aliases=True), fg='green', bold=True)


bdk.add_command(run)
bdk.add_command(project)


def main():
    bdk(None)


if __name__ == "__main__":
    main()



from __future__ import annotations
from badook_tests.dsl.summary import Summary
from badook_tests.dsl.checks import Calculation


class TimeSeriesAnalysisSummary(Summary):
    def __init__(self, features: str, name: str, window_size: int, n_mads_allowed: float, time_column: str = None):
        super().__init__(features, name)
        if not isinstance(features, str):
            raise Exception(
                "Feature name for Hampel filter must be a string")
        self.type = 'TimeSeriesAnalysisSummary'
        self.window_size = window_size
        self.n_mads_allowed = n_mads_allowed
        self.time_column = time_column
        self.partitions = None
        self.agg = None

    @property
    def value(self):
        return Calculation(self.data, 'value', self._ctx)

    @property
    def window_rolling_median(self):
        return Calculation(self.data, 'window_rolling_median', self._ctx)

    @property
    def window_rolling_mad(self):
        return Calculation(self.data, 'window_rolling_MAD', self._ctx)

    @property
    def window_lower_boundary(self):
        return Calculation(self.data, 'window_lower_boundary', self._ctx)

    @property
    def window_upper_boundary(self):
        return Calculation(self.data, 'window_upper_boundary', self._ctx)

    @property
    def is_outlier(self):
        return Calculation(self.data, 'is_outlier', self._ctx)

    def partition_by(self, *args) -> Summary:
        self.partitions = list(args)
        return self

    def set_agg(self, agg: str) -> Summary:
        self.agg = agg
        return self



from __future__ import annotations
from badook_tests.dsl.summary import Summary
from badook_tests.dsl.checks import Calculation


class Pearson(Summary):
    def __init__(self, features: list, name: str):
        super().__init__(features, name)
        if len(features) < 2:
            raise Exception(
                "Feature list in correlation summary must include at least two features.")
        self.method = 'pearson'
        self.type = 'CorrelationSummary'

    @property
    def r_coeff(self):
        return Calculation(self.data, 'r_coeff', self._ctx)



import shortuuid


class Check(object):

    def __init__(self, from_dataset, calc_name, check_fn, ctx, to_dataset=None, to_calc_name=None, join_keys=None):

        self.id = shortuuid.uuid()
        self.project_id = ctx.project_name
        self.calc_name = calc_name
        self.from_dataset = from_dataset

        if not to_dataset is None:
            self.to_dataset = to_dataset
            if not to_calc_name is None:
                self.to_calc_name = to_calc_name

        if join_keys:
            self.join_keys = join_keys

        self.check_fn = check_fn
        self.test_name = ctx.test_name
        self._ctx = ctx

    def assert_with_tolerance(self, tolerance: int = 0) -> None:
        if self._ctx is None:
            raise ValueError(
                "context is not set, make sure your summery was added to a summary table correctly")
        if tolerance is not None:
            self.tolerance = tolerance
        self._ctx.assert_check(self)
        self._ctx.run_mgr.add_waiting_execution(
            f'running check for {self.test_name}', self.id)

    def assert_all(self) -> None:
        self.assert_with_tolerance()


class Calculation(object):

    def __init__(self, from_dataset, name, ctx):
        self.from_dataset = from_dataset
        self.name = name
        self._ctx = ctx
        self.to_calc_name = None

    def compare_to(self, summary, calculation=None):
        if not calculation and not summary:
            return None
        if 'parent' in summary.__dict__:
            summary = ToSummaryData(summary.dataset, summary.name, summary.parent, summary.run_id,
                                    summary.groups, summary.time_window, summary.type)
        self.to_dataset = summary
        if calculation:
            self.to_calc_name = calculation

        return self

    def join_on(self, on=None, keys_l=None, keys_r=None):
        """
        on, keys_l, keys_r: Can be string with single column name, or list with several columns name
        Examples:
            1. join_on('Suburb')
            2. join_on(['Suburb','Type'])
            3. join_on(keys_l = 'Suburb', keys_r = 'Suburb_copy')
            4. join_on(keys_l = ['Suburb', 'Rooms'], keys_r = ['Suburb_copy', 'Rooms']

        :return: join_keys = {'keys_l': keys_l, 'keys_r': keys_r}
        """
        if on:
            if type(on) is str:
                on = [on]
            self.join_keys = {
                'keys_l': on,
                'keys_r': on
            }

        elif keys_l and keys_r:
            if type(keys_l) is str:
                keys_l = [keys_l]
            if type(keys_r) is str:
                keys_r = [keys_r]

            self.join_keys = {
                'keys_l': keys_l,
                'keys_r': keys_r
            }
        return self

    def check(self, check_fn):
        if hasattr(self, 'to_dataset'):
            if hasattr(self, 'join_keys'):
                return Check(self.from_dataset, self.name, check_fn, self._ctx, self.to_dataset, self.to_calc_name, self.join_keys)
            else:
                return Check(self.from_dataset, self.name, check_fn, self._ctx, self.to_dataset, self.to_calc_name)

        return Check(self.from_dataset, self.name, check_fn, self._ctx)


class ToSummaryData(object):

    def __init__(self, dataset, summary_name, name, run_id, join_on, time_window, summary_type):
        self.name = name
        self.summary_name = summary_name
        self.run_id = run_id
        self.join_on = join_on
        self.time_window = time_window
        self.dataset = dataset
        self.type = summary_type



from urllib import *
import json
import hmac
import base64
import hashlib
import requests
import logging
import datetime
import dateutil.tz
import urllib.parse
from requests.auth import AuthBase

logger = logging.getLogger(__name__)
logger.disabled = True


class Client(object):

    def __init__(self, api_key, secret_key, base_uri):
        self.auth = HmacAuthWrapper(api_key, secret_key)
        self._base_uri = base_uri

    def send_post_request(self, data, path=''):
        try:
            result = requests.post(urllib.parse.urljoin(self._base_uri, path), data=data, auth=self.auth, headers={
                'Content-Type': 'application/json'})

            return self.handle_request_ressult(result, 'send_post_request')

        except Exception as e:
            return self.handle_request_exception(e, path)

    # returns None if response is not 200

    def send_get_request(self, params={}, path=''):
        try:
            result = requests.get(urllib.parse.urljoin(self._base_uri, path), params=params, auth=self.auth, headers={
                'Content-Type': 'application/json'})

            return self.handle_request_ressult(result, 'send_get_request')

        except Exception as e:
            return self.handle_request_exception(e, path, 'send_get_request')

    def send_json_post_request(self, data, path=''):
        json_data = json.dumps(data)
        return self.send_post_request(json_data, path)

    def send_delete_request(self, data='', path=''):
        try:
            result = requests.delete(urllib.parse.urljoin(self._base_uri, path), data=data, auth=self.auth, headers={
                'Content-Type': 'application/json'})

            return self.handle_request_ressult(result, 'send_delete_request')

        except Exception as e:
            return self.handle_request_exception(e, path, 'send_delete_request')

    def send_json_delete_request(self, data, path=''):
        json_data = json.dumps(data)
        return self.send_delete_request(data=json_data, path=path)

    def handle_request_ressult(self, result, request_name='request'):
        success = (result.status_code >= 200 and result.status_code < 300)
        error = self.convert_request_error_dict(result, success)

        response = {
            'error': error,
            'data': None
        }

        # handle warnings
        if result.status_code == 199:
            logger.warning(
                f"{request_name} - {result.status_code} response received for {result.url} - {error['message']}")
            response['error'] = False

        # handle errors
        elif result.status_code >= 400:
            logger.error(
                f"{request_name} - {result.status_code} response received for {result.url} - {error['message']}")

            if result.status_code == 404:
                response['error']['message'] = '404 - Not found'

        # handle success
        elif success:
            response['error'] = False

            try:
                response['data'] = result.json()

            except Exception:
                # if response content is plain text, no need to pass aything to data
                pass

        '''
        response {
            error: { message: Str, user: Boolean } OR False,
            data: {}
        }
        '''
        return response

    def handle_request_exception(self, ex, path, request_name='request'):
        logger.warning(
            f"{request_name} - Exception {ex} raised for {path}")

        response = {
            'error': {
                'message': 'Could not connect to server: check your network connection',
                'user': False
            },
            'data': None
        }
        return response

    def convert_request_error_dict(self, result, success):

        # convert request's error to an object with this structure:
        error = {
            'message': '',
            'user': False
        }

        try:
            error = result.json()

            # if the result is an error (and it's content is not a plain text)
            # make sure it has 'message' and 'user' fields
            if not success:
                # check if has a 'message' key
                if 'message' not in error:
                    error['message'] = ''
                    error['user'] = False
                elif 'user' not in error:
                    error['user'] = False

        except Exception:
            # meaning the message is plain text
            error = {
                'message': result.text,
                'user': False
            }
            pass

        return error


class HmacAuthWrapper(AuthBase):
    DELIMITER = '\n'

    def __init__(self, api_key, secret_key):
        self.api_key = api_key
        self.secret_key = secret_key

    def __call__(self, request):
        self._encode(request)
        return request

    def _encode(self, request):
        timestamp = datetime.datetime.now(dateutil.tz.tzutc()).isoformat()
        self._add_api_key(request)
        self._add_signature(request, timestamp)
        request.headers['X-Auth-Timestamp'] = timestamp
        request.headers['X-Auth-Version'] = '1'

    def _add_api_key(self, request):
        request.headers['X-API-Key'] = self.api_key
        request.headers['X-Auth-Path'] = request.path_url

    def _add_signature(self, request, timestamp):
        method = request.method
        path = request.path_url
        content = '' if request.body is None else request.body
        signature = self._sign(method, timestamp, path, content)
        request.headers['X-Auth-Signature'] = signature

    def _sign(self, method, timestamp, path, content):
        # Build the message to sign
        message = bytes(method, 'utf-8') + \
            bytes(self.DELIMITER, 'utf-8') + \
            bytes(timestamp, 'utf-8') + \
            bytes(self.DELIMITER, 'utf-8') + \
            bytes(path, 'utf-8') + \
            bytes(self.DELIMITER, 'utf-8') + \
            bytes(content, 'utf-8')

        # Create the signature
        digest = hmac.new(bytes(self.secret_key, 'utf-8'),
                          message, digestmod=hashlib.sha256).digest()
        res = base64.b64encode(digest).decode('utf-8')
        return res



import asyncio
import json
import logging
import signal
import sys
import threading
import unittest

import click
import emoji
import pysher
from badook_tests.config import get_config
from badook_tests.util import Halo

from badook_tests.context import global_context
from badook_tests.context.context import BadookContext

logger = logging.getLogger(__name__)
logger.disabled = True


# Exception classes
class TestFailedError(Exception):
    """Raised when a test execution fails"""
    pass


# classes
class DataTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        config = get_config()
        DataTestCase.context = BadookContext(
            config.client_id, config.client_secret)
        #signal.signal(signal.SIGINT, cls.signal_handler)
        loop = asyncio.get_event_loop()

        # run customer's set_up function to create project/dataset summaries
        cls.set_up()

        # try:
        run_mgr = cls.init_run()
        cls.context.run_mgr = run_mgr

        # check if sumarries hase identical display names - get the first one found
        duplicate_names = cls.get_summaries_name_duplication(
            cls.context.dataset_summaries.items())

        if not bool(duplicate_names):
            cls.handle_summaries(run_mgr, loop)
        else:
            names = ', '.join(duplicate_names)
            err_message = f"Summary name/s: {names}, has been found in multiple summaries.\nPlease change to a unique name for each summary and try again."
            logger.error(err_message)
            raise Exception({'message': err_message})

        # except Exception as e:
        #     terminate = cls.handle_exception(e.args[0])

        #     if terminate:
        #         # send "run failed" to mgmt API
        #         cls.send_fail_to_api()
        #         assert False, ""

    @classmethod
    def init_run(cls):
        proj = cls.context.project_name

        # registering a new test run with the badook mgmt API.
        # if this is a rerun-by-id request, then the mgmt API has to "redo" an existing run and delete all data ("RerunFull")
        if global_context.rerun:

            # send request to service manager to update the run with 'rerun'
            response = cls.context.api_client.send_json_post_request(
                {'projectId': proj, 'runType': 'RerunFull',
                 'runId': global_context.run_id}, "runhistory/run")

            if response['error']:
                error_msg = f"Re-Run - {response['error']['message']}"
                logger.error(error_msg)
                raise Exception(error_msg)

            cls.run_id = global_context.run_id

        # not a rerun - the mgmt API returns our new RunID which we'll use for all DS summaries
        else:
            # send request to service manager to create a new "run"
            response = cls.start_run(proj)

            if response['error']:
                error_msg = f"Run - {response['error']['message']}"
                logger.error(error_msg)
                raise Exception(error_msg)

            cls.run_id = response['data']['runId']  # pylint: disable=E1136

        cls.context.set_run_id(cls.run_id)
        cls.channel_id = response['data']['orgId']  # pylint: disable=E1136

        click.secho(
            f'\nrunning project - {proj}, run id: {cls.run_id}', fg='magenta', bold=True)

        res = RunManager(cls.channel_id, cls.run_id)
        return res

    @classmethod
    def start_run(cls, proj):
        '''
        This is a utility method for starting a run with the management API.
        Its main use is to be mocked in tests. You can patch the methods as follows:

        @patch('badook_tests.data_test_case.DataTestCase.start_run', return_value=[the value you want to be returned])
        '''
        return cls.context.api_client.send_json_post_request(
            {'projectId': proj}, "runhistory/run")

    @classmethod
    def set_up(self):
        "Hook method for setting up the test fixture before exercising it."
        pass

    def setUp(self):
        self.context = self.__class__.context
        self.context.test_name = self._testMethodName

        self.__dict__.update(self.context.dataset_summaries)
        self.run_id = self.__class__.run_id
        self.org_id = self.__class__.channel_id

    @classmethod
    def handle_summaries(cls, run_mgr, loop):

        futures = []
        for name, item in cls.context.dataset_summaries.items():
            display_name = item['display_name']
            ds = item['dataset_summary']

            if global_context.rerun:
                # this means we want to rerun a specific batch that's already run with this run_id for this dataset.

                # get batch data correspondence to main dataset and join-on if exists by run id
                response = cls.get_batch_data_by_run_id(
                    ds, global_context.run_id)

                request_data = {
                    'rerun': True,
                    'project_id': cls.context.project_name,
                    'run_id': [global_context.run_id]
                }
                del_response = cls.context.client.send_json_delete_request(
                    request_data, path=f'run')

                if del_response['error']:
                    raise Exception(del_response['error']['message'])

            else:
                # get latest batched data correspondence to main dataset and join-on if exists
                response = cls.get_next_unexecuted_batch_data(ds)

            if not response['error']:
                main_ds_batch = cls.handle_get_batch_data_response(
                    ds, response)
                text = f'{display_name} on batch {main_ds_batch}' if main_ds_batch else display_name

                if global_context.rerun:
                    batch_name = f' {main_ds_batch}' if main_ds_batch else 'es'
                    print(f're-running batch{batch_name}\n')

                futures.append(asyncio.ensure_future(
                    run_mgr.add_execution(f'collecting dataset: {text}', name)))

                if global_context.force_pull:
                    ds.force_pull = True
            else:
                err_message = f"Batches - {response['error']['message']}"
                logger.error(err_message)
                raise Exception({'message': err_message})

            ds.set_project_id(cls.context.project_name)
            ds.set_run_id(cls.run_id)
            ds.collect()

        # this gathers all the dataset futures into a *concurrent* future so they'd run at the same time
        dataset_futures = asyncio.gather(*futures)

        # this calls the runloop with a function that waits on the parallel futures and the error future for the first to complete
        loop.run_until_complete(run_mgr.run_summaries([asyncio.ensure_future(
            run_mgr.add_error_execution()), dataset_futures]))

        super().run

    @classmethod
    def get_summaries_name_duplication(cls, summaryList):
        def get_display_name(obj):
            (_, item) = obj
            return item['display_name']

        duplicate_names = []
        display_name_list = list(
            map(lambda obj: get_display_name(obj), summaryList))

        for display_name in display_name_list:
            if (display_name_list.count(display_name) > 1) and not (duplicate_names.__contains__(display_name)):
                duplicate_names.append(display_name)

        return_list = list(map(lambda name: f'"{name}"', duplicate_names))
        return return_list

    @classmethod
    def get_next_unexecuted_batch_data(cls, ds):
        # this service will get last batch file for each dataset (in case there also join-on datasets)
        """
        example:
        ret_response = {
            error: {'message': '', 'user': False}
            data = {
                dataset : {
                    name: <dataset>,
                    batch_value: <batch>,
                }
                join-on : [{
                    name: <dataset1>,
                    batch_value: <batch>
                },{
                    name: <dataset2>,
                    batch_value: <batch>
                }]
                ...
            }
        }
        """
        data = {}
        project_id = cls.context.project_name

        # get all dataset names that need to extract their batches list
        dataset_ids = [ds.dataset]  # main dataset

        # add additional datasets if exist in join-on
        if len(ds.joined_datasets) > 0:
            for key in ds.joined_datasets:
                dataset_ids.append(key)

        # get batches list for each dataset
        for dataset_id in dataset_ids:
            path = f'batch/{cls.channel_id}/{dataset_id}/{project_id}'
            response = cls.context.client.send_get_request(path=path)

            if not response['error']:
                batch = None
                if response['data'] is not None:
                    batch = cls.get_latest_batch(response['data'])
                data[dataset_id] = {'batch_value': batch}
            else:
                # stop the loop and return with the current error response
                return response

        return {'data': cls.convert_batch_data_to_response_data(data), 'error': False}

    @classmethod
    def get_batch_data_by_run_id(cls, ds, run_id):
        # this service will get specific run batch file for each dataset (in case there also join-on datasets)
        """
        example:
        ret_response = {
            error: {'message': '', 'user': False}
            data = {
                dataset : {
                    name: <dataset>,
                    batch_value: <batch>,
                }
                join-on : [{
                    name: <dataset1>,
                    batch_value: <batch>
                },{
                    name: <dataset2>,
                    batch_value: <batch>
                }]
                ...
            }
        }
        """
        data = {}
        project_id = cls.context.project_name

        # get all dataset names that need to extract their batches list
        dataset_ids = [ds.dataset]  # main dataset

        # add additional datasets if exist in join-on
        if len(ds.joined_datasets) > 0:
            for key in ds.joined_datasets:
                dataset_ids.append(key)

        # get batches list for each dataset
        for dataset_id in dataset_ids:

            # dig up the batch name by run id
            path = f'batch/{cls.channel_id}/{dataset_id}/{project_id}/{run_id}'
            response = cls.context.client.send_get_request(path=path)

            if not response['error']:
                batch = None

                # if readBatch flag is False the response will return False (which doesn't have 'get')
                if isinstance(response['data'], dict):
                    batch = response['data'].get('batch_value', None)
                data[dataset_id] = {'batch_value': batch}
            else:
                # stop the loop and return with the current error response
                err_message = f"Batch by run id - {response['error']['message']}"
                raise Exception(err_message)
                # return response

        return {'data': cls.convert_batch_data_to_response_data(data), 'error': False}

    @classmethod
    def handle_get_batch_data_response(cls, ds, response):

        data = response['data']
        main_ds_batch = data['dataset']['batch_value']  # main dataset batch

        # if there is a batch to run, add "batch_value" field in ds
        if main_ds_batch:
            ds.set_batch(main_ds_batch)

        if data['join_on']:
            for item_data in data['join_on']:
                item_data_name = item_data['name']
                item_data_batch = item_data['batch_value']

                # if there is a batch to run, add "batch_value" field in ds
                if item_data_batch:
                    ds.joined_datasets[item_data_name]['batch_value'] = item_data['batch_value']

        return main_ds_batch

    @classmethod
    def get_latest_batch(cls, data):

        # sort batches by created date
        content = sorted(data, key=lambda item: item['created_date'])

        batches = [b for b in content if b['batch_value']]
        batch = batches.pop(0)['batch_value'] if batches else None
        return batch

    @classmethod
    def convert_batch_data_to_response_data(self, data):
        """
        data:
            dataset1: {batch_value} -> main batch
            dataset2: {batch_value} -> join-on batch
            dataset3: {batch_value} -> join-on batch
            ....

        response_data:{
              data = {
                dataset : {
                    name: <dataset>,
                    batch_value: <batch>,
                }
                join-on : [{
                    name: <dataset1>,
                    batch_value: <batch>
                },{
                    name: <dataset2>,
                    batch_value: <batch>
                }]
                ...
            }
        }
        """
        response_data = {}
        if len(data) > 0:
            response_data['join_on'] = []

        for index, key in enumerate(data, start=0):
            data_item = {'name': key, 'batch_value': data[key]['batch_value']}

            # main dataset
            if index == 0:
                response_data['dataset'] = data_item

            # join-on datasets
            else:
                response_data['join_on'].append(data_item)

        return response_data

    @classmethod
    def tearDownClass(cls):
        data = {"orgId": cls.channel_id,
                "projectId": cls.context.project_name, "runId": cls.run_id}
        cls.context.api_client.send_json_post_request(
            data, "runhistory/run/end")

    # @classmethod
    # def signal_handler(cls, sig, frame):
    #     click.secho(emoji.emojize(
    #         '\n\nOh, no! You have aborted the tests! :scream:', use_aliases=True), fg='red', bold=True)

    #     cls.send_fail_to_api()
    #     sys.exit(0)

    @classmethod
    def send_fail_to_api(cls):
        data = {"orgId": cls.channel_id, "projectId": cls.context.project_name,
                "runId": cls.run_id, "status": "Failure"}
        cls.context.api_client.send_json_post_request(
            data, "runhistory/run/end")

    @classmethod
    def handle_exception(cls, error):
        # exception can be either plain text or a dict {'message': text}
        # plain text will be displayed with the generic error notice, and
        # the dict's message will be displayed on it's own.
        generic = True
        terminate = True
        msg = error

        # check if argument is a dict type
        if isinstance(error, dict):
            # check if it has a 'message' element
            msg = error.get('message', None)
            terminate = error.get('terminate', True)
            if msg is not None:
                generic = False

        if (generic):
            click.secho(
                f'\nUh Oh! Something bad happened while trying to run the test. Please contact Badook support.\n({msg})',
                fg='red', bold=True)
        else:
            click.secho(
                f'\n{msg}\n', fg='red', bold=True)

        return terminate


class RunManager(object):

    def __init__(self, channel_id, run_id) -> None:
        root = logging.getLogger('pysher.connection')
        root.disabled = True

        self._channel_id = channel_id
        self._run_id = run_id
        self.pusher = pysher.Pusher('515d671ab968b85e463a', cluster='ap1')
        self.pusher.connection.bind(
            'pusher:connection_established', self.pysher_connection)

        self.pusher.connect()

        self._lock = threading.Lock()
        self.loop = asyncio.get_event_loop()
        self._events = {}
        self._error_event = {}

    # waits for the first future to complete and then goes over the result to make sure exceptions are thrown
    async def run_summaries(self, futures):
        (done, _) = await asyncio.wait(futures, return_when='FIRST_COMPLETED')
        for t in done:
            t.result()

    def summary_event_handler(self, data):
        event_data = json.loads(data)
        dataset_name = event_data['dataset_summary']
        result = event_data['status']
        error = event_data.get('error', False)

        if dataset_name:
            event = self._events.get(dataset_name, False)
            if event:
                event['result'] = result
                event['error'] = event_data.get('error', False)
                event['warning'] = event_data.get('warning', False)
                self.loop.call_soon_threadsafe(event['event'].set)

        elif result == 'Failure' and error and not dataset_name:
            # meaning the error occured before could identify which dataset it is
            # therefore, terminate the whole process
            raise Exception(error)

    def test_event_handler(self, data):
        check_data = json.loads(data)

        # in case the event hasn't been added to self._events yet
        check_id = check_data['check_id']
        event = self._events.get(check_id, False)

        if event:
            event['result'] = check_data['status']
            event['error_message'] = check_data['error_message']
            self.loop.call_soon_threadsafe(event['event'].set)

    def error_event_handler(self, data):
        # This error event is called to close the whole run proccess
        error_data = json.loads(data)

        # get list of all running datasets events
        running_datasets = self._events.keys()

        # loop on all current events and set to close them with Failure
        for dataset in running_datasets:
            event = self._events[dataset]
            event['result'] = 'Failure'
            self.loop.call_soon_threadsafe(event['event'].set)

        # set error event with error message
        error_message = 'Badook backend error'

        if error_data['error_message']:
            error_message = error_data['error_message']

        self._error_event['error_message'] = error_message
        self.loop.call_soon_threadsafe(self._error_event['event'].set)

    def get_error_in_event(self, error):
        error_message = 'Badook backend error'
        error_type = error['error_type']

        if error_type and error_type.lower() == 'user':
            error_message = {'message': error['error_message']}
            error_message['terminate'] = error['terminate']

        return error_message

    def pysher_connection(self, data):
        channel = self.pusher.subscribe(f'badook-{self._channel_id}')
        channel.bind(f'{self._run_id}-summary', self.summary_event_handler)
        channel.bind(f'{self._run_id}-test', self.test_event_handler)
        channel.bind(f'{self._run_id}-error', self.error_event_handler)

    async def add_execution(self, msg: str, id: str) -> None:
        ev = self.create_event(id)
        await self.print_summary_execution(msg, ev)

    async def add_error_execution(self) -> None:
        self._error_event['event'] = asyncio.Event()
        await self._error_event['event'].wait()
        raise Exception(self._error_event['error_message'])

    def add_waiting_execution(self, msg: str, id: str, spinner='circleQuarters', color='cyan') -> None:
        ev = self.create_event(f'{id}')
        self.loop.run_until_complete(self.print_test_execution(msg, ev))

    async def print_test_execution(self, msg: str, ev, spinner='circleQuarters', color="cyan",
                                   text_color="cyan") -> None:
        with Halo(text=msg, spinner=spinner, color=color, text_color=text_color) as spinner:
            await ev['event'].wait()
            if ev['result'].lower() == 'success':
                spinner.succeed()  # "✔"
            else:
                spinner.fail()  # "✖"
                raise TestFailedError(ev['error_message'])

    async def print_summary_execution(self, msg: str, ev, spinner='dots', color="green", text_color="green") -> None:
        with Halo(text=msg, spinner=spinner, color=color, text_color=text_color) as spinner:
            await ev['event'].wait()
            error = False

            # handle spinner's icons
            if ev['result'].lower() == 'success':
                spinner.succeed()  # "✔"

            else:
                spinner.fail()  # "✖"
                error = ev.get('error', False)

            # if the run has a warning, display it
            if ev['warning']:
                click.secho(
                    f'\nWarnings:\n{ev["warning"]}', fg='yellow', bold=True)

            # if the run has an error, raise an exception
            if error:
                raise Exception(self.get_error_in_event(error))

    def create_event(self, id: str):
        with self._lock:
            self._events[id] = {'event': asyncio.Event()}
            return self._events[id]



from badook_tests.dsl.summary import DatasetSummary, Summary
import base64
import json
import logging
import urllib.parse

import click
import cloudpickle

from badook_tests.config import get_config
from badook_tests.context.client import Client
from badook_tests.dsl.checks import Check
from badook_tests.dsl.enums import ComparisonPoint
from badook_tests.dsl.summary import DatasetSummary, Summary

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.DEBUG)


class BadookContext(object):

    client = None

    def __init__(self, client_id, client_secret):
        config = get_config()
        self.client = Client(client_id, client_secret, urllib.parse.urljoin(
            config.data_cluster_url, 'orchestrator/'))
        self.api_client = Client(
            client_id, client_secret, config.management_cluster_url)
        self.dataset_summaries = {}

    def collect(self, data, path=''):
        request_data = json.dumps(data, cls=RequestEncoder)
        logger.debug('Summary data: {}'.format(request_data))
        response = self.client.send_post_request(request_data, path)

        if response['error']:
            raise Exception(
                f"collect summaries - {response['error']['message']}")

    def create_summary(self, name, dataset) -> DatasetSummary:

        dataset_id = self._validate_dataset(dataset)
        dataset_summary = DatasetSummary(self)
        dataset_summary.set_dataset(dataset_id)
        dataset_summary.set_name(name)
        self.dataset_summaries[dataset_summary.name] = {
            'display_name': dataset_summary.display_name,
            'dataset_summary': dataset_summary
        }

        return dataset_summary

    def assert_check(self, check):
        check_data = json.dumps(check, cls=CheckEncoder)
        logger.debug('check data: %s', check_data)
        response = self.client.send_post_request(check_data, 'tests')
        if response['error']:
            raise Exception(f"assert check - {response['error']['message']}")

    def set_project_name(self, name):
        self.project_name = name

    def set_run_id(self, run_id):
        self.run_id = run_id
        for ds in self.dataset_summaries.values():
            ds['dataset_summary'].set_run_id(run_id)
            for summ in ds['dataset_summary'].summaries.values():
                summ.set_run_id(run_id)

    def _validate_dataset(self, dataset):
        response = self.api_client.send_json_post_request(
            {'datasetName': dataset}, 'dataset/validate/name')
        error = response.get('error', False)
        if error:
            click.secho(f"Error: {error['message']}", fg='red', bold=True)
            raise Exception()

        return response['data']['id']  # pylint: disable=E1136


class RequestEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, DatasetSummary):
            x = obj.__dict__.copy()
            x['summaries'] = [self.ser_summaries(
                summ, obj.summaries[summ]) for summ in obj.summaries]
            del x["_ctx"]

            # serializing the user feature functions
            if bool(obj.user_features):
                features = {}
                for feature in obj.user_features:
                    features.update(self.ser_udcs(
                        feature, obj.user_features[feature]))
                x['user_features'] = features

            return x

        if isinstance(obj, ComparisonPoint):
            return str(obj).split(".")[1]
        else:
            return

    def ser_summaries(self, name: str, summ: Summary):
        res = summ.__dict__.copy()
        del res["_ctx"]

        # removing the claculation attribute before serialization
        for clac in summ.user_calculations:
            del res[clac]

        # serializing the user calculation functions
        if bool(summ.user_calculations):
            calcs = {}
            for fn in summ.user_calculations:
                calcs.update(self.ser_udcs(fn, summ.user_calculations[fn]))
            res['user_calculations'] = calcs
        return res

    def ser_udcs(self, name: str, udc):
        res = {}
        res[name] = {'func': base64.encodebytes(
            cloudpickle.dumps(udc['func'], protocol=4)).decode('utf-8'),
            'input_columns': udc['input_columns']
        }
        return res


class CheckEncoder (json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, Check):

            # copy.deepcopy(obj.__dict__)
            result = {key: value for key,
                      value in obj.__dict__.items() if key != "_ctx"}
            fn = base64.encodebytes(cloudpickle.dumps(
                obj.check_fn, protocol=4)).decode('utf-8')

            result['from_dataset'] = obj.from_dataset.__dict__

            if hasattr(obj, 'to_dataset'):
                result['to_dataset'] = obj.to_dataset.__dict__

            result['check_fn'] = fn
            return result
        else:
            return



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['badook_tests',
 'badook_tests.cli',
 'badook_tests.config',
 'badook_tests.context',
 'badook_tests.dsl',
 'badook_tests.dsl.Cluster',
 'badook_tests.dsl.CorrelationSummary',
 'badook_tests.dsl.enums',
 'badook_tests.dsl.summaries',
 'badook_tests.util']

package_data = \
{'': ['*']}

install_requires = \
['Pysher>=1.0.6,<2.0.0',
 'click>=8.0.1,<9.0.0',
 'cloudpickle>=1.6,<3.0',
 'emoji>=1.4.2,<2.0.0',
 'halo>=0.0.31,<0.0.32',
 'ipython>=7.27.0,<8.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'requests>=2.26.0,<3.0.0',
 'shortuuid>=1.0.1,<2.0.0',
 'websocket-client==0.57.0']

entry_points = \
{'console_scripts': ['bdk = badook_tests.cli.bdk:main']}

setup_kwargs = {
    'name': 'badook-tests',
    'version': '0.2.0rc0',
    'description': 'badook data testing framework for Python',
    'long_description': '# badook tests python SDK\n\n## Setup\n\n### Prerequisites\n\nCurrent version supports Python 3.7 only\n\n### Installation\nTo install badook from pip use:\n\n```\npython -m pip install badook-tests\n```\n\n## Running the example localy\n\nTo run using a local server first set the local address correctly in the `config\\badook.yaml` file under the `data_cluster_url` entry.\nNext run the example using the following command:\n\n```{python}\npython examples/test_dsl_example.py\n```\n',
    'author': 'badook Engineering',
    'author_email': 'engineering@badook.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/badook-ai/badook-tests-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<3.8.0',
}


setup(**setup_kwargs)



from __future__ import annotations
from badook_tests.dsl.summary import Summary
from badook_tests.dsl.checks import Calculation


class LabelSummary(Summary):
    def __init__(self, features: list, name: str):
        super().__init__(features, name)
        if len(features) < 2:
            raise Exception(
                "Feature list in label summary must include at least two features.")
        self.type = 'LabelSummary'

    @property
    def avg_agreement(self):
        return Calculation(self.data, 'avg_agreement', self._ctx)

    @property
    def median_agreement(self):
        return Calculation(self.data, 'median_agreement', self._ctx)

    @property
    def fleiss_or_cohens_kappa(self):
        return Calculation(self.data, 'fleiss_or_cohens_kappa', self._ctx)



import click
import unittest
from badook_tests import BadookRuntimeTestResult
from badook_tests.context import global_context


@click.command()
@click.option(
    "--target-directory",
    "-d",
    default="./",
    help="The root of the project directory where you want to execute badook tests.",
)
@click.option('--batch/--no-batch', default=False)
@click.option('--force-pull/--no-force', default=False)
@click.option('--rerun-id')
@click.option('--file_path', '-f', default='')
def run(target_directory, batch, force_pull, rerun_id, file_path):
    global_context.track_batch = batch
    global_context.force_pull = force_pull

    pattern = 'test*.py'

    if file_path:
        if not file_path.endswith('.py'):
            file_path += '.py'

        split_paths = file_path.rsplit('/', 1)

        if len(split_paths) > 1:
            target_directory, pattern = split_paths
        else:
            target_directory = ''
            pattern = split_paths[0]

        click.secho(f'file={pattern}', fg='blue')

    suit = unittest.defaultTestLoader.discover(
        target_directory, pattern=pattern)
    click.secho(f'found {len(suit._tests)} test cases', fg='green')
    if rerun_id:
        click.secho(f'rerun id={rerun_id}', fg='blue')
        global_context.run_id = rerun_id
        global_context.rerun = True

    unittest.TextTestRunner(resultclass=BadookRuntimeTestResult).run(suit)



import getpass

import click
from badook_tests.config import get_config

from badook_tests.context import BadookContext


@click.command()
@click.option('--delete', multiple=True)
@click.option('--delete-run', multiple=True)
@click.option('--from-project', multiple=False)
def project(delete, delete_run, from_project):
    if delete:
        project_list = ', '.join(delete)
        click.secho(f'deleting projects: {project_list}', fg='cyan', bold=True)
        send_delete_project_request(list(delete))

    elif delete_run and from_project:
        send_delete_runs_from_project_request(from_project, list(delete_run))


def send_delete_runs_from_project_request(project_id, runs):
    path = 'runhistory/run'
    config = get_config()
    context = BadookContext(config.client_id, config.client_secret)

    # send service manager request to delet projects
    body = {"projectId": project_id, "runIds": runs}
    response = context.api_client.send_json_delete_request(body, path)

    # send orchestrator request to delete projects records from postgres tables
    if not response['error']:
        request_data = {
            'project_id': project_id,
            'run_id': runs
        }
        response = context.client.send_json_delete_request(
            request_data, path=f'run')

    handle_response(response)


def send_delete_project_request(projects):
    path = f"runhistory/project"

    config = get_config()
    context = BadookContext(config.client_id, config.client_secret)

    # send service manager request to delete projects
    response = context.api_client.send_json_delete_request(
        {'projects': projects}, path)

    # send orchestrator request to delete projects records from postgres tables
    if not response['error']:
        request_data = {
            'project_id': projects
        }
        response = context.client.send_json_delete_request(
            request_data, path=f'project')

    handle_response(response)


def handle_response(response):
    if response['error']:
        err_msg = response['error']['message']
        click.secho(f'\nERROR: {err_msg}', fg='red', bold=True)
    else:
        click.secho(f'\ndeleting projects was successful',
                    fg='green', bold=True)
